# Modified from https://github.com/fnpy/fn.py
# Refactor later

from .func import F, curry
from .array import Stream, Seq, AbstractSeq
from .underscore import shortcut as _
from .args import Args as __
from . import monad
from .monad import return_option, Option, Some, Empty

__all__ = ["F", "Stream", "_", "curry", "monad", "__", "return_option", "Option", "Some", "Empty", "AbstractSeq", "Seq"]
